# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2011 Akretion LDTA (<http://www.akretion.com>).
#    @author Raphaël Valyi
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from tools.translate import _
from osv import fields, osv
from datetime import datetime
import netsvc


class project_product_allocation(osv.osv):
    _name = 'project.product.allocation'
    _description = 'Project Product Allocation'
    _rec_name = 'product_id'
    
    _columns = {
        'product_id': fields.many2one('product.product', 'Product', required=True),
        'product_estimated_qty': fields.float('Quantity'),
        'product_corrected_qty': fields.float('Corrected Quantity'),
        'sale_flow': fields.selection([('normal', 'Normal'), ('normal_mto', 'Normal MTO'), ('direct_delivery', 'Direct Delivery'), ('direct_invoice', 'Direct Invoice/Indirect Delivery'), ('direct_invoice_and_delivery', 'Direct Invoice')], 'Sale Flow', help="Is this order tied to a sale order? How will it be delivered and invoiced then?"),
        'is_on_service_invoice': fields.boolean('On Service Invoice?', help='True if invoiced along with service'),
        'phase_id': fields.many2one('project.phase', 'Project Phase', ondelete='cascade', required=True),
        'sale_id' : fields.many2one("sale.order.line", "Order Line"),
        'checked': fields.boolean('Checked', help='Utility checkbox, use as you want'),
    }
    _defaults = {
        'sale_flow': 'normal',
        'is_on_service_invoice': True,
        'product_estimated_qty': 0,
    }

project_product_allocation()

class stock_picking(osv.osv):
    _inherit = "stock.picking"

    _columns = {
        'phase_id': fields.many2one('project.phase', 'Project Phase', ondelete='set null'),
        'related_project_id':fields.related('phase_id', 'project_id', type='many2one', relation='project.project', string='Project', store=True),
    }

stock_picking()

class project_phase(osv.osv):
    _inherit = "project.phase"

    _columns = {
        'product_ids': fields.one2many('project.product.allocation', 'phase_id', "Phase Products"),
        'picking_ids': fields.one2many('stock.picking', 'phase_id', "Phase Pickings"),
    }
    
    def set_open(self, cr, uid, ids, *args): # Check list to set open
        res = super(project_phase, self).set_open(cr, uid, ids, args)
        for phase in self.browse(cr, uid, ids):
            # Check that at least one picking is in "done" state if there are pickings
            picked = False
            for picking in phase.picking_ids:
                if picking.state in ('done'):
                    picked = True
                    break
            if not picked and phase.picking_ids:
                raise osv.except_osv(_('Warning!'), _('At least one picking should be done'))
            
            # Check that the phase is ordered if there are products
            if phase.product_ids:
                order_line_ids = self.pool.get('sale.order.line').search(cr, uid, [('phase_id', '=', phase.id)])
                if order_line_ids:
                    valid_order = False
                    for order_line in self.pool.get("sale.order.line").browse(cr, uid, order_line_ids):
                        if order_line.order_id.state not in ('draft', 'cancel'):
                            valid_order = True
                            break
                    if not valid_order:
                        raise osv.except_osv(_('Warning!'), _('None of the related phase sale orders are in a valid state!'))
                else:
                    raise osv.except_osv(_('Warning!'), _('This phase hasn\'t been ordered yet'))

        return res


    def get_or_create_purchase(self, cr, uid, order, phase, purchases, sale_flow, is_on_service_invoice, product_id, dest_address_id, context):
        supplier_data = self.pool.get('product.product').read(cr, uid, product_id, ['seller_id'])['seller_id']
        if supplier_data:
            supplier_id = supplier_data #FIXME supplier_data[0] ?
        else:
            raise osv.except_osv('No supplier information!', "product %s has no supplier information!" % self.pool.get('product.product').read(cr, uid, product_id, ['name'])['name'])
        if purchases.get((sale_flow, is_on_service_invoice, supplier_id, dest_address_id), False):
            return purchases[(sale_flow, is_on_service_invoice, supplier_id, dest_address_id)]
        else:
            supplier_data = self.pool.get('purchase.order').onchange_partner_id(cr, uid, [], supplier_id)['value']
            partner_address_id = supplier_data['partner_address_id']
            pricelist_id = supplier_data['pricelist_id']
            fiscal_position = supplier_data['fiscal_position']#TODO override for Brazil!! and make modular!
            warehouse_id = self.pool.get('purchase.order').default_get(cr, uid, ['warehouse_id'], context)['warehouse_id']
            sale_flow_data = self.pool.get('purchase.order').sale_flow_change(cr, uid, [], sale_flow, supplier_id, warehouse_id)['value']
            invoice_method = sale_flow_data.get('invoice_method', 'picking')
            location_id = sale_flow_data['location_id']
            if sale_flow in ['direct_delivery', 'direct_invoice_and_delivery']:
                location_id = phase.project_id.location_id.id
            purchase_id = self.pool.get('purchase.order').create(cr, uid, {
                      'analytic_account_id': phase.project_id.analytic_account_id.id,
                      'partner_id': supplier_id,
                      'partner_address_id': partner_address_id,
                      'pricelist_id': pricelist_id,
                      'sale_flow': sale_flow != 'normal_mto' and sale_flow or 'normal',
                      'location_id': location_id,
                      'invoice_method': invoice_method,
                      'dest_address_id': dest_address_id,
                      'origin': phase.project_id.name,
                      'phase_id': phase.id,
                      'sale_id': order.id,
                      })
            purchases[(sale_flow, is_on_service_invoice, supplier_id, dest_address_id)] = purchase_id
            return purchase_id


    def get_or_create_picking(self, cr, uid, order, phase, pickings, sale_flow, is_on_service_invoice, context):
        if pickings.get((sale_flow, is_on_service_invoice), False):
            return pickings[(sale_flow, is_on_service_invoice)]
        else:
            picking_id = self.pool.get('stock.picking').create(cr, uid, {
                     'name': phase.name,
                     'sale_id': order.id,
                     'address_id': order.partner_shipping_id.id,
                     'note': order.note,
                     'origin': "project " + phase.project_id.name + " - " + phase.name,
                     'type': 'out',
                     'invoice_state': (sale_flow != 'direct_invoice') and '2binvoiced' or 'none',
                     'company_id': phase.project_id.company_id.id,
                     'phase_id': phase.id
            })
            pickings[(sale_flow, is_on_service_invoice)] = picking_id
            return picking_id


    def add_order_line(self, cr, uid, order_id, project_product_allocation, phase):
        date_planned = phase.date_start or datetime.now()
        order_data = self.pool.get('purchase.order').read(cr, uid, order_id, ['pricelist_id', 'partner_id', 'fiscal_position'])
        pricelist_id = order_data['pricelist_id'][0]
        supplier_id = order_data['partner_id'][0]
        fiscal_position_id = order_data['fiscal_position'] and order_data['fiscal_position'][0] or False
        order_line_data = self.pool.get('purchase.order.line').product_id_change(cr, uid, [], pricelist_id, project_product_allocation.product_id.id, project_product_allocation.product_estimated_qty, False, supplier_id, date_planned, fiscal_position_id , date_planned, False, False, phase.project_id.name + " - " + phase.name)['value']
        if project_product_allocation.sale_flow == 'direct_delivery':
            account_analytic_id = phase.project_id.analytic_account_id.id
        else:
            account_analytic_id = False
        purchase_order_line_id = self.pool.get('purchase.order.line').create(cr, uid, {
                'product_id': project_product_allocation.product_id.id,
                'order_id': order_id,
                'product_uom': order_line_data['product_uom'],
                'price_unit': order_line_data['price_unit'],
                'name': order_line_data['name'],
                'product_qty': order_line_data['product_qty'],
                'date_planned': order_line_data['date_planned'],
                'account_analytic_id': account_analytic_id
        })
        return purchase_order_line_id


    def add_procurement(self, cr, uid, picking_id, project_product_allocation, phase, order, proc_state='draft', purchase_id=False):
        date_planned = phase.date_start or datetime.now()
        line = project_product_allocation.sale_id
        move_id = self.pool.get('stock.move').create(cr, uid, {
                'name': project_product_allocation.product_id.name,
                'picking_id': picking_id,
                'product_id': project_product_allocation.product_id.id,
                'date_planned': date_planned,
                'date_expected': date_planned,
                'product_qty': project_product_allocation.product_corrected_qty or project_product_allocation.product_estimated_qty,
                'product_uom': project_product_allocation.product_id.uom_id.id,
                'product_uos_qty': project_product_allocation.product_corrected_qty or project_product_allocation.product_estimated_qty,
                'product_uos': project_product_allocation.product_id.uom_id.id,
                #'product_packaging': line.product_packaging.id,
                #'address_id': line.address_allotment_id.id or order.partner_shipping_id.id,
                'location_id': phase.project_id.warehouse_id.lot_stock_id.id,
                'location_dest_id': phase.project_id.location_id.id,
                'sale_line_id': line.id,
                #'tracking_id': False,
                'state': 'draft',
                #'state': 'waiting',
                #'note': line.notes,
                'company_id': phase.project_id.company_id.id,
            })
    
        proc_id = self.pool.get('procurement.order').create(cr, uid, {
                'name': line.name,
                'origin': order.name + " - " + phase.project_id.name + " - " + phase.name,
                'date_planned': date_planned,
                'product_id': line.product_id.id,
                'product_qty': line.product_uom_qty,
                'product_uom': line.product_id.uom_id.id,
                'product_uos_qty': (line.product_uos and line.product_uos_qty)\
                        or line.product_uom_qty,
                'product_uos': (line.product_uos and line.product_uos.id)\
                        or line.product_uom.id,
                'location_id': phase.project_id.warehouse_id.lot_stock_id.id,
                'procure_method': line.type,
                'move_id': move_id,
                'property_ids': [(6, 0, [x.id for x in line.property_ids])],
                'company_id': phase.project_id.company_id.id,
                'purchase_id': purchase_id,
            })
        self.pool.get('procurement.order').write(cr, uid, proc_id, {'state': proc_state})
        return proc_id

        
    
    def action_create_picking(self, cr, uid, ids, context=None):
        res = {}
        for phase in self.browse(cr, uid, ids): #TODO deal if picking exists already eventually
            if phase.product_ids:
                if not phase.project_id.warehouse_id or not phase.project_id.location_id:
                    raise osv.except_osv(_('Data Insufficient !'), _('In order to plan the material expeditions, you should define source and target locations for this project'))
                if phase.product_ids and not phase.date_start:
                    raise osv.except_osv(_('Data Insufficient !'), _('In order to plan the material requisitions, you need to define a start date for phase: %s!') % (phase.name,))
                can_create_picking = True
                for picking in phase.picking_ids:
                    if picking.state != "cancel":
                        can_create_picking = False #only create picking if none is there already.
                if can_create_picking:
                    proc_ids = []

                    order = False

                    pickings = {}
                    purchases = {}


                    for project_product_allocation in phase.product_ids:
                        if project_product_allocation.sale_id:#TODO because of that condition picking might be empty, check before!
                            line = project_product_allocation.sale_id
                            order = line.order_id

                            purchase_order_id = False
                            proc_state = 'draft'
                            if project_product_allocation.sale_flow != 'normal' and line.product_id.supply_method == 'buy': #then it's a buy on order!
                                sale_flow = project_product_allocation.sale_flow == 'normal_mto' and 'normal' or project_product_allocation.sale_flow
                                purchase_order_id = self.get_or_create_purchase(cr, uid, order, phase, purchases, sale_flow, project_product_allocation.is_on_service_invoice, line.product_id.id, phase.project_id.location_id.id, context)
                                self.add_order_line(cr, uid, purchase_order_id, project_product_allocation, phase)
                                proc_state = 'running'

                            if project_product_allocation.sale_flow in ['normal', 'normal_mto', 'direct_invoice']:
                                picking_id = self.get_or_create_picking(cr, uid, order, phase, pickings, project_product_allocation.sale_flow, project_product_allocation.is_on_service_invoice, context)
                                proc_id = self.add_procurement(cr, uid, picking_id, project_product_allocation, phase, order, proc_state, purchase_order_id)
                                proc_ids.append(proc_id)
                                self.pool.get('sale.order.line').write(cr, uid, [line.id], {'procurement_id': proc_id})


                    if not order:
                        raise osv.except_osv(_('Error!'), _('You should first sale at least some products of phase %s, or alternatively create a picking manually!') % (phase.name,))
                    val = {}
                    for (k, picking_id) in pickings.items():
                        wf_service = netsvc.LocalService("workflow")
                        wf_service.trg_validate(uid, 'stock.picking', picking_id, 'button_confirm', cr)
        
                    for proc_id in proc_ids:
                        wf_service = netsvc.LocalService("workflow")
                        wf_service.trg_validate(uid, 'procurement.order', proc_id, 'button_confirm', cr)
                    
#                                if order.state == 'shipping_except':
#                                    val['state'] = 'progress'
#                    
#                                    if (order.order_policy == 'manual'):
#                                        for line in order.order_line:
#                                            if (not line.invoiced) and (line.state not in ('cancel', 'draft')):
#                                                val['state'] = 'manual'
#                                                break
#                                self.write(cr, uid, [order.id], val)
        
        return True#res #TODO return something useful

project_phase()


class project_project(osv.osv):
    _inherit = "project.project"

    _columns = {
        'location_id': fields.many2one('stock.location', 'Stock Location'),
        'warehouse_id': fields.many2one('stock.warehouse', 'Main Warehouse'),
    }

#    def validate_project(self, cr, uid, ids, context=None):
#        res = super(project_project, self).validate_project(cr, uid, ids, context)
#        self.action_create_pickings(cr, uid, ids, context)
#        return res
    
    def action_create_pickings(self, cr, uid, ids, context=None):
        res = {}
        for project in self.browse(cr, uid, ids): #TODO deal if picking exists already eventually
            res[project.id] = self.pool.get('project.phase').action_create_picking(cr, uid, self.pool.get('project.project').read(cr, uid, project.id, ['phase_ids'])['phase_ids'], context)
        return True#res #TODO return something useful
        
    def create(self, cr, uid, vals, ctx=None):
        if not ctx.get('partner_shipping_id', False) or vals.get('location_id', False):
            return super(project_project, self).create(cr, uid, vals, ctx)
        ids = self.pool.get('stock.location').search(cr, uid, [('address_id', '=', ctx.get('partner_shipping_id'))])
        if ids:
            vals.update({'location_id': ids[0]})
        else:
            location_id = self.pool.get('res.partner.address').browse(cr, uid, ctx['partner_shipping_id'], ctx).partner_id.property_stock_customer.id
            id = self.pool.get('stock.location').create(cr, uid, {'name': vals.get('name', False), 'location_id': location_id, 'address_id': ctx.get('partner_shipping_id'), 'usage': 'customer'}, ctx)
            vals.update({'location_id': id})
        return super(project_project, self).create(cr, uid, vals, ctx)
    
    def get_order_lines(self, cr, uid, ids, context=None):
        order_lines = super(project_project, self).get_order_lines(cr, uid, ids, context)
        for project in self.browse(cr, uid, ids):
            for phase in project.phase_ids:
                for project_product_allocation in phase.product_ids:
                    if not project_product_allocation.sale_id:
                        if project_product_allocation.product_estimated_qty < 1:
                            raise osv.except_osv(_('Data Insufficient !'), _('Please check the Quantity in Project Product Allocations, it should not be less than 1!'))
                        order_lines.append({
                                           'line': {'phase_id': phase.id,
                                            'resource_product_id': project_product_allocation.id,
                                            'product_id': project_product_allocation.product_id.id,
                                            'product_uom_qty': project_product_allocation.product_estimated_qty,
                                            'product_uom': project_product_allocation.product_id.uom_id.id,
                                            'type': project_product_allocation.sale_flow == 'normal' and 'make_to_stock' or 'make_to_order'
                                            },
                                            'resource_model': 'project.product.allocation',
                                            'resource_id': project_product_allocation.id
                                            })
        return order_lines

project_project()

class stock_picking(osv.osv):
    _inherit = "stock.picking"

    def action_done(self, cr, uid, ids, context=None): #TODO is it the bettest place to put that code?
        for picking in self.browse(cr, uid, ids):
            for move in picking.move_lines:
                is_pure_mts = True
                for proc in move.procurements:
                    print proc.procure_method
                    if proc.procure_method == "make_to_order":
                        is_pure_mts = False #FIXME if move is partially MTS we do nothing today. What are the implications?
                        break
                print move.sale_line_id, move.sale_line_id and move.sale_line_id.order_id
                if is_pure_mts and move.sale_line_id and move.sale_line_id.order_id.project_id:# and move.sale_line_id.order_id.project_id:
                    sign = picking.type == 'in' and 1 or -1
                    vals = {
                        'name': picking.name,
                        'journal_id': self.pool.get('account.analytic.journal').search(cr, uid, [('code', '=', 'DELIV')])[0],
                        'account_id': move.sale_line_id.order_id.project_id.id,
                        'general_account_id': sign == 1 and (move.product_id.property_account_expense and move.product_id.property_account_expense.id or move.product_id.categ_id.property_account_expense_categ.id) or (move.product_id.property_account_income and move.product_id.property_account_income.id or move.product_id.categ_id.property_account_income_categ.id),
                        'amount': sign * move.product_qty * move.product_id.standard_price,
                        'date': datetime.now(), #TODO possibility to specify phase?
                        }
                    self.pool.get("account.analytic.line").create(cr, uid, vals, context)
        return super(stock_picking, self).action_done(cr, uid, ids, context)

    def action_cancel(self, cr, uid, ids, context=None):
#        TODO find and cancel analytic move:
        return super(stock_picking, self).action_cancel(cr, uid, ids, context)

stock_picking()


class procurement_order(osv.osv):
    _inherit = 'procurement.order'

    _columns = {
        'phase_id': fields.many2one('project.phase', 'Project Phase'),
        'related_project_id':fields.related('phase_id', 'project_id', type='many2one', relation='project.project', string='Project', store=True),
    }

    def make_po(self, cr, uid, ids, context=None):
        ids_for_super = []
        for procurement in self.browse(cr, uid, ids, context=context):
            if procurement.purchase_id:
                self.write(cr, uid, [procurement.id], {'state': 'running'})
            else:
                ids_for_super.append(procurement.id)
        return super(procurement_order, self).make_po(cr, uid, ids_for_super, context)

procurement_order()
